(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * FWTesting.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************.***************************************************************)
(* $Id: Section5Example.thy 9105 2011-01-25 22:03:54Z makarius $ *)

theory NATNormalisationTest
imports FWTesting
begin


text {* 
An example of how to test using the distributivity of the normalisation.
*}


declare [[testgen_profiling]]

definition internet::"IntegerPort net" where "internet \<equiv> {{(a,b). a = 7}}"

definition dmz::"IntegerPort net" where "dmz \<equiv> {{(a,b). a= 6}}"

definition intranet::"IntegerPort net" where "intranet \<equiv> {{(a,b). a= 5}}"

definition intranet2::"IntegerPort net" where "intranet2 \<equiv> {{(a,b). a= 25}}"


definition Filter :: "(IntegerPort net, port) Combinators" where 
"Filter \<equiv> DenyAll \<oplus>
          AllowPortFromTo intranet internet 8 \<oplus>
          AllowPortFromTo intranet2 internet 8 \<oplus>
          AllowPortFromTo intranet dmz 443 \<oplus>
          AllowPortFromTo dmz intranet 25 \<oplus>
          AllowPortFromTo intranet dmz 25 \<oplus>
          AllowPortFromTo internet dmz 380 \<oplus>
          AllowPortFromTo dmz intranet 325 \<oplus>
          AllowPortFromTo intranet dmz 525 \<oplus>
          AllowPortFromTo internet dmz 80 \<oplus>
          AllowPortFromTo intranet2 dmz 25\<oplus>
          AllowPortFromTo intranet dmz 5235 \<oplus>
          AllowPortFromTo internet dmz 820 \<oplus>
          AllowPortFromTo internet dmz 265\<oplus>
          AllowPortFromTo intranet dmz 5225 \<oplus>
          AllowPortFromTo internet dmz 810 \<oplus>
          AllowPortFromTo internet dmz 525\<oplus>
          AllowPortFromTo intranet dmz 5825 \<oplus>
          AllowPortFromTo internet dmz 380 \<oplus>
          AllowPortFromTo internet dmz 425"



definition NAT1  ::"(IntegerPort,'b) packet \<mapsto> 
     (((IntegerPort,'b) packet) set)" where
 "NAT1 = pat2pool_int {7} {17,27,37}"


definition NAT2  ::"(IntegerPort,DummyContent) packet \<mapsto> 
          (((IntegerPort,DummyContent) packet) set)" where
 "NAT2 = pat2pool_int {6} {16,26,36}"

definition NATID :: "(IntegerPort,DummyContent) packet \<mapsto> 
        (((IntegerPort,DummyContent) packet) set)" where
 "NATID = (\<lambda>p. Some (allow {p}))"

definition NAT :: "(IntegerPort,DummyContent) packet \<mapsto> 
        (((IntegerPort,DummyContent) packet) set)" where
 "NAT = (NAT1 \<Oplus> NAT2 \<Oplus> NATID)"


definition Policy ::"(IntegerPort,DummyContent) packet \<mapsto> 
        (((IntegerPort,DummyContent) packet) set)" where
 "Policy = ((\<lambda> (x,y). x) o_f (NAT \<Otimes>\<^sub>2 (C Filter) o (\<lambda>x. (x,x))))"



lemmas NAT = NAT2_def NATID_def NAT1_def



fun not_equal_aux :: "nat \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool" where
 "not_equal_aux (Suc n) x y = ((((x!n) \<noteq> (y!n)) \<and> ((y!n)\<noteq>(x!n))) \<and> not_equal_aux n x y)"
|"not_equal_aux Zero x y = True"


definition not_equal where
 "not_equal x = not_equal_aux (Suc (length x)) x  x"

lemma not_equals: "\<not> not_equal [internet,intranet,dmz]"
apply (simp add: not_equal_def)
done



lemma nets_different: "internet \<noteq> intranet \<and> internet \<noteq> dmz \<and> intranet \<noteq> dmz \<and>
                       intranet \<noteq> internet \<and> dmz \<noteq> internet \<and> dmz \<noteq> intranet \<and> 
                       internet \<noteq> intranet2 \<and> intranet2 \<noteq> dmz \<and> 
                       intranet \<noteq> intranet2 \<and> intranet2 \<noteq> intranet \<and> 
                       intranet2 \<noteq> internet \<and> dmz \<noteq> internet \<and> dmz \<noteq> intranet2"
apply (simp add: internet_def intranet_def dmz_def intranet2_def)
apply (simp add: sets_distinct1 sets_distinct2 sets_distinct5 sets_distinct6)
done




lemma noMT: "\<forall> x \<in> set (policy2list Filter). dom (C x) \<noteq> {}" 
apply (simp add: Filter_def)
apply (simp add: if_splits PLemmas internet_def intranet_def dmz_def intranet2_def
  IntegerPort.src_port)
done



lemma ANDex: "allNetsDistinct (policy2list Filter)"
apply (simp add: Filter_def allNetsDistinct_def nets_different)
apply (auto simp: PLemmas internet_def intranet_def dmz_def intranet2_def netsDistinct_def
  IntegerPort.src_port)
done


lemma all_in_list: "all_in_list (policy2list Filter) (Nets_List Filter)"
apply (simp add: Filter_def)
apply (unfold Nets_List_def)
apply (unfold bothNets_def)
apply (rule conjI)
apply (insert nets_different)
apply simp
apply (simp add: aux2)
apply (simp add: nets_different)
apply auto
done



lemma C_Filter_compile: " 
C (list2policy (normalize Filter)) =  C Filter"
apply (unfold normalize_def)
apply (rule C_eq_compile)
apply (simp add: Filter_def)
apply (rule all_in_list)
apply (rule ANDex)
done


lemma norm_essential1: "{a,b} = {b,a}"
  apply auto
done


lemma norm_essential2: "\<lbrakk>x \<noteq> a; y\<noteq>b; (x \<noteq> y \<and> x \<noteq> b) \<or>  (a \<noteq> b \<and> a \<noteq> y)\<rbrakk> \<Longrightarrow> {x,a} \<noteq> {y,b}"
  apply auto
done





lemma normFilter: "normalize Filter = [DenyAll,
        DenyAllFromTo NATNormalisationTest.intranet NATNormalisationTest.internet \<oplus>
        DenyAllFromTo NATNormalisationTest.internet NATNormalisationTest.intranet \<oplus>
        AllowPortFromTo NATNormalisationTest.intranet NATNormalisationTest.internet 8,
        DenyAllFromTo intranet2 NATNormalisationTest.internet \<oplus>
        DenyAllFromTo NATNormalisationTest.internet intranet2 \<oplus>
        AllowPortFromTo intranet2 NATNormalisationTest.internet 8,
        DenyAllFromTo intranet2 NATNormalisationTest.dmz \<oplus>
        DenyAllFromTo NATNormalisationTest.dmz intranet2 \<oplus>
        AllowPortFromTo intranet2 NATNormalisationTest.dmz 25,
        DenyAllFromTo NATNormalisationTest.intranet NATNormalisationTest.dmz \<oplus>
        DenyAllFromTo NATNormalisationTest.dmz NATNormalisationTest.intranet \<oplus>
        ((((((AllowPortFromTo NATNormalisationTest.intranet NATNormalisationTest.dmz 443 \<oplus>
              AllowPortFromTo NATNormalisationTest.dmz NATNormalisationTest.intranet 25) \<oplus>
             AllowPortFromTo NATNormalisationTest.intranet NATNormalisationTest.dmz 25) \<oplus>
            AllowPortFromTo NATNormalisationTest.dmz NATNormalisationTest.intranet 325) \<oplus>
           AllowPortFromTo NATNormalisationTest.intranet NATNormalisationTest.dmz 525) \<oplus>
          AllowPortFromTo NATNormalisationTest.intranet NATNormalisationTest.dmz 5235) \<oplus>
         AllowPortFromTo NATNormalisationTest.intranet NATNormalisationTest.dmz 5225) \<oplus>
        AllowPortFromTo NATNormalisationTest.intranet NATNormalisationTest.dmz 5825,
        DenyAllFromTo NATNormalisationTest.internet NATNormalisationTest.dmz \<oplus>
        DenyAllFromTo NATNormalisationTest.dmz NATNormalisationTest.internet \<oplus>
        (((((AllowPortFromTo NATNormalisationTest.internet NATNormalisationTest.dmz 80 \<oplus>
             AllowPortFromTo NATNormalisationTest.internet NATNormalisationTest.dmz 820) \<oplus>
            AllowPortFromTo NATNormalisationTest.internet NATNormalisationTest.dmz 265) \<oplus>
           AllowPortFromTo NATNormalisationTest.internet NATNormalisationTest.dmz 810) \<oplus>
          AllowPortFromTo NATNormalisationTest.internet NATNormalisationTest.dmz 525) \<oplus>
         AllowPortFromTo NATNormalisationTest.internet NATNormalisationTest.dmz 380) \<oplus>
        AllowPortFromTo NATNormalisationTest.internet NATNormalisationTest.dmz 425]"
apply (insert nets_different noMT)
apply (simp add:  normalize_def Filter_def bothNets_def norm_essential1 
                  norm_essential2 Nets_List_def)
done

lemma NATuniv: "dom NAT = UNIV"
apply (simp add: dom_def NAT_def NAT)
done


lemma dom_2_subset1: "dom (r o_f ((A \<Otimes>\<^sub>2 B) o (\<lambda> x. (x,x)))) \<subseteq> dom A"
apply (simp add: dom_def prod_2_def o_apply policy_range_comp_def)
apply (simp split: option.splits decision.splits)
apply auto
done

lemma dom_2_comm: "dom (A \<Otimes>\<^sub>2 B) = dom (B \<Otimes>\<^sub>2 A)"
apply (simp add: dom_def prod_2_def o_apply policy_range_comp_def)
apply (simp split: option.splits decision.splits prod.splits)
oops



lemma dom_2_subset2: "dom (r o_f ((A \<Otimes>\<^sub>2 B) o (\<lambda> x. (x,x)))) \<subseteq> dom B"
apply (rule subsetI)
apply (simp add: dom_def prod_2_def o_apply policy_range_comp_def)
apply (simp split: option.splits decision.splits)
done


lemma dom_2: "dom A = UNIV \<Longrightarrow>
 dom (r o_f ((A \<Otimes>\<^sub>2 B) o (\<lambda> x. (x,x)))) = dom B"
apply (rule equalityI)
apply (rule dom_2_subset2)
apply (rule subsetI)
apply (simp add: dom_def prod_2_def o_apply policy_range_comp_def)
apply (simp split: option.splits decision.splits)
apply auto
done

lemma domDA1: "dom (C (DenyAll \<oplus> P)) = UNIV"
apply (simp add: PLemmas)
apply (auto simp: deny_all_def C.simps)
apply (auto split: option.splits simp: deny_all_def)
done


lemma PolUniv: "dom Policy = UNIV"
apply (simp add: Policy_def)
apply (simp add: Filter_def)
apply (insert NATuniv)
apply (simp add: dom_2)
apply (rule domDA1)
done


lemma if_None: "((if (\<forall>ac. (\<exists>bc. fst (snd (a, (aa, b), (ab, ba), bb)) = (ac, bc)) \<longrightarrow> ac = 7) \<and>
            (\<forall>ac. (\<exists>bc. fst (snd (snd (a, (aa, b), (ab, ba), bb))) = (ac, bc)) \<longrightarrow> ac = 6)
         then deny_all (a, (aa, b), (ab, ba), bb) else None)
 =
       None) = Y"
apply simp
oops



test_spec "FUT x = Policy x"
apply (prepare_fw_spec)
apply (rule_tac PL = "(NAT \<Otimes>\<^sub>L (Comb2PolList (rev (normalize (Filter))) C)) 
                       (op \<Otimes>\<^sub>2) (\<lambda> (x,y). x) (\<lambda> x. (x,x))" 
   in distrFUTLG1)
apply (simp add: UNIV_I PolUniv)
apply (simp add: Policy_def)
apply (rule sym)
apply (subst normalizeNAT)
apply (simp add: Filter_def)
apply (rule ANDex)
apply (rule all_in_list)
apply simp
apply (simp add: gatherDomain_def normFilter FUTListGD_def)
apply (insert NATuniv)
apply (simp add: dom_2 domain_reasoning)
apply (thin_tac "?X")
apply (simp add: PLemmas intranet_def internet_def C.simps 
      map_add_def dmz_def split_tupled_all intranet2_def option.splits
      prod_2_id_def prod_2_def NAT NAT_def pat2pool_int_def  if_splits
      policy_range_comp_def id_def NetworkCore.id_def prod.splits
Filter_def      content_def allow_all_def deny_all_def dest_def content_def) 
apply (tactic  
 "REPEAT(CHANGED(TRYALL((rtac @{thm conjI}) THEN' (rtac @{thm impI}))))")      
apply (simp add: PLemmas intranet_def internet_def C.simps 
      map_add_def dmz_def split_tupled_all intranet2_def option.splits
      prod_2_id_def prod_2_def NAT NAT_def pat2pool_int_def  if_splits
      policy_range_comp_def id_def NetworkCore.id_def prod.splits
      content_def allow_all_def deny_all_def dest_def content_def)
apply (simp split: option.splits)
apply (simp add: if_splits split:decision.splits)

apply (gen_test_cases FUT simp: deny_all_def allow_all_def PLemmas)
apply (simp add: PLemmas deny_all_def)
apply (simp add: PLemmas intranet_def internet_def C.simps 
      map_add_def dmz_def split_tupled_all intranet2_def 
      prod_2_id_def prod_2_def NAT NAT_def pat2pool_int_def  
      policy_range_comp_def id_def NetworkCore.id_def
      content_def allow_all_def deny_all_def dest_def content_def)
apply (simp add: PO_def)
store_test_thm "foo"

gen_test_data "foo"

thm foo.test_data

thm foo.pos


test_spec "FUT x = Policy x"
apply (prepare_fw_spec)
apply (simp_all add: PLemmas intranet_def intranet2_def internet_def 
      Filter_def C.simps  map_add_def dmz_def split_tupled_all
      prod_2_id_def prod_2_def NAT NAT_def pat2pool_int_def  
      policy_range_comp_def id_def intranet2_def  allow_all_def deny_all_def
      content_def Policy_def NetworkCore.id_def)
apply (gen_test_cases FUT simp: deny_all_def allow_all_def)

store_test_thm "bar"

gen_test_data "bar"

thm bar.test_data

thm bar.pos


end
