theory Normalisation
imports UPFCore
begin


text{* This theory provides the formalisations required for the
normalisation of UPF policies. A typical usage scenario can be
observed in the firewall case study.

The scenario is the following (can be applied iteratively):

 - Two policies are combined using one of the parallel combinators
 - (e.g. P = P1 \<Otimes>\<^sub>\<or>\<^sub>D P2) (At least) one of the constituent policies has
 - a normalisation procedures, which as output produces a list of
 - policies that are semantically equivalent to the original policy if
 - combined from left to right using the override operator.

*}


fun prod_list :: "('\<alpha> \<mapsto>'\<beta>) \<Rightarrow> (('\<gamma> \<mapsto>'\<delta>) list) \<Rightarrow> 
                  (('\<alpha> \<mapsto>'\<beta>) \<Rightarrow> ('\<gamma> \<mapsto>'\<delta>) \<Rightarrow> (('\<alpha> \<times> '\<gamma>) \<mapsto> ('\<beta> \<times> '\<delta>))) \<Rightarrow>
                  (('\<beta> \<times> '\<delta>) \<Rightarrow> 'y) \<Rightarrow> ('x \<Rightarrow> ('\<alpha> \<times> '\<gamma>)) \<Rightarrow>  
                  (('x \<mapsto> 'y) list)"  (infixr "\<Otimes>\<^sub>L" 54) where
  "prod_list x (y#ys) par_comb ran_adapt dom_adapt = 
         ((ran_adapt o_f ((par_comb x y) o dom_adapt))#(prod_list x ys par_comb ran_adapt dom_adapt))"
| "prod_list x [] par_comb ran_adapt dom_adapt = []"



text{* An instance *}

definition prod_2_list :: "[('\<alpha> \<mapsto>'\<beta>), (('\<gamma> \<mapsto>'\<delta>) list)] \<Rightarrow> 
                  (('\<beta> \<times> '\<delta>) \<Rightarrow> 'y) \<Rightarrow> ('x \<Rightarrow> ('\<alpha> \<times> '\<gamma>)) \<Rightarrow> 
                  (('x \<mapsto> 'y) list)" (infixr "\<Otimes>\<^sub>2\<^sub>L" 55) where 
  "x \<Otimes>\<^sub>2\<^sub>L y =  (\<lambda> d r. (x \<Otimes>\<^sub>L y) (op \<Otimes>\<^sub>2) d r)"  


fun Comb2PolList::"'a list \<Rightarrow> ('a \<Rightarrow> ('b \<mapsto> 'c)) \<Rightarrow> 
                   (('b \<mapsto> 'c) list)" 
where
 "Comb2PolList (x#y) sem = ((sem x)#(Comb2PolList y sem))"
|"Comb2PolList [] sem = []"


fun list2policyP::"('a \<mapsto> 'b) list \<Rightarrow>
                   ('a \<mapsto> 'b)" 
where
 "list2policyP (x#y) = (x \<Oplus> (list2policyP y))"
|"list2policyP [] = empty"


lemma list2listNMT[rule_format]: "x \<noteq> [] \<longrightarrow> Comb2PolList x sem \<noteq> []"
apply (case_tac x)
apply simp_all
done
 
lemma two_conc: "(prod_list x (y#ys) p r d)=  
((r o_f ((p x y) o d))#(prod_list x ys p r d))"
by simp



definition is_distr where
 "is_distr p = (\<lambda> g f. (\<forall> N P1 P2. ((g o_f ((p N (P1 \<Oplus> P2)) o f)) = 
               ((g o_f ((p N P1) o f)) \<Oplus> (g o_f ((p N P2)  o f))))))"


lemma "r o_f (list2policyP ((x \<Otimes>\<^sub>L y) p id id)) o d = (list2policyP ((x \<Otimes>\<^sub>L y) p r d))"
apply (induct_tac y)
apply simp_all
apply (simp add: policy_range_comp_def)
apply (simp add: policy_range_comp_def)
apply (rule ext)+
apply (simp split: option.split decision.splits)
oops


definition is_strict where
 "is_strict p = (\<lambda> r d. \<forall> P1. (r o_f (p P1 \<emptyset> \<circ> d)) = \<emptyset>)"


lemma is_distr_orD: "is_distr (op \<Otimes>\<^sub>\<or>\<^sub>D) d r"
apply (simp add: is_distr_def)
apply (rule allI)+
apply (rule distr_orD)
by simp


lemma is_strict_orD: "is_strict (op \<Otimes>\<^sub>\<or>\<^sub>D) d r"
apply (simp add: is_strict_def)
apply (simp add: policy_range_comp_def)
done


lemma is_distr_2: "is_distr (op \<Otimes>\<^sub>2) d r"
apply (simp add: is_distr_def)
apply (rule allI)+
apply (rule distr_or2)
by simp

lemma is_strict_2: "is_strict (op \<Otimes>\<^sub>2) d r"
apply (simp only: is_strict_def)
apply simp
apply (simp add: policy_range_comp_def)
done


lemma domStart: "t \<in> dom p1 \<Longrightarrow> (p1 \<Oplus> p2) t = p1 t"
apply (simp add: map_add_dom_app_simps)
done


lemma notDom: "x \<in> dom A \<Longrightarrow> \<not> A x = None"
apply auto
done


declare Fun.o_apply [simp del]

lemma aux[rule_format]: 
 "x \<in> dom (r o_f (A \<circ> d)) \<longrightarrow>
    (r o_f  (A \<Oplus> B) o d) x = (r o_f (A \<circ> d)) x"
by (simp add: prod_2_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)



lemma aux2[rule_format]: "x \<in> dom (r o_f ((P1 \<Otimes>\<^sub>2 a) \<circ> d)) \<longrightarrow>
         ((r o_f ((P1 \<Otimes>\<^sub>2 a \<Oplus> Q) \<circ> d)) x = (r o_f ((P1 \<Otimes>\<^sub>2 a) \<circ> d)) x) "
apply (simp add: prod_2_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done


lemma aux3a[rule_format]: 
 "x \<notin> dom ( (P1 \<Otimes>\<^sub>2 a \<circ> d)) \<longrightarrow>
  ((((P1 \<Otimes>\<^sub>2 a \<Oplus> list2policyP P2 \<circ> d)) x) =
  ((((P1 \<Otimes>\<^sub>2 list2policyP P2) \<circ> d)) x))"
apply (simp add: prod_2_def dom_def policy_range_comp_def map_add_def o_apply)
apply (simp add: prod_2_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done

lemma aux3[rule_format]: 
 "x \<notin> dom (r o_f (P1 \<Otimes>\<^sub>2 a \<circ> d)) \<longrightarrow>
  (((r o_f (P1 \<Otimes>\<^sub>2 a \<Oplus> list2policyP P2 \<circ> d)) x) =
  ((r o_f  ((P1 \<Otimes>\<^sub>2 list2policyP P2) \<circ> d)) x))"
apply (simp add: prod_2_def dom_def policy_range_comp_def map_add_def o_apply)
apply (simp add: prod_2_def dom_def policy_range_comp_def map_add_def o_apply
         split: option.splits decision.splits prod.splits)
done



lemma Norm_Distr[rule_format]: 
 "((r o_f (((op \<Otimes>\<^sub>2) P1 (list2policyP P2)) o d)) x = 
  ((list2policyP ((P1 \<Otimes>\<^sub>L P2) (op \<Otimes>\<^sub>2) r d)) x))"
apply (induct P2) 
apply (simp add: policy_range_comp_def o_apply prod_2_def)
apply simp_all
apply (case_tac "x \<in> dom (r o_f (P1 \<Otimes>\<^sub>2 a \<circ> d))")
apply (subst domStart) back
apply simp_all
apply (simp add: aux2)
apply (simp add: aux3)
by (metis map_add_dom_app_simps(3) prod_2_list_def)


text {* Some domain reasoning *}



lemma domSubsetNAT: "dom A = UNIV \<Longrightarrow> dom ((\<lambda>(x, y). x) o_f (A \<Otimes>\<^sub>2 B) o (\<lambda> x. (x,x))) = dom B"
apply (rule set_eqI)
apply (rule iffI)
by (auto simp: prod_2_def policy_range_comp_def dom_def o_apply
              split: decision.splits option.splits prod.splits)

declare o_apply [simp]


end


