(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * PolicyCombinators.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: PolicyCombinators.thy 9105 2011-01-25 22:03:54Z makarius $ *)

theory PolicyCombinators
imports 
PolicyCore
begin

text{* In order to ease the specification of a concrete policy, we
  define some combinators. Using these combinators, the specification
  of a policy gets very easy, and can be done similarly as in tools
  like IPTables.  *}

definition 
  allow_all    :: "(('\<alpha>,'\<beta>) packet \<mapsto> unit)" where 
  "allow_all p  = Some (allow ())" 

definition 
  deny_all :: "(('\<alpha>,'\<beta>) packet \<mapsto> unit)" where
  "deny_all p   = Some (deny ())"              

definition 
  allow_all_from :: "'\<alpha>::adr net \<Rightarrow> (('\<alpha>,'\<beta>) packet \<mapsto> unit)" where
  "allow_all_from src_net = allow_all |` {pa. src pa \<sqsubset> src_net}" 

definition 
  deny_all_from    :: "'\<alpha>::adr net \<Rightarrow> (('\<alpha>,'\<beta>) packet \<mapsto> unit)" where
  "deny_all_from src_net   = deny_all |` {pa. src pa \<sqsubset>  src_net}"
        
definition
  allow_all_to    :: "'\<alpha>::adr net \<Rightarrow> (('\<alpha>,'\<beta>) packet \<mapsto> unit)" where
  "allow_all_to dest_net   = allow_all |` {pa. dest pa \<sqsubset>  dest_net}"

definition 
  deny_all_to :: "'\<alpha>::adr net \<Rightarrow> (('\<alpha>,'\<beta>) packet \<mapsto> unit)" where
  "deny_all_to dest_net = deny_all |` {pa. dest pa \<sqsubset> dest_net}"

definition 
  allow_all_from_to    :: "'\<alpha>::adr net \<Rightarrow> '\<alpha>::adr net \<Rightarrow> (('\<alpha>,'\<beta>) packet \<mapsto> unit)" where
  "allow_all_from_to src_net dest_net = allow_all |`
                                   {pa. src pa \<sqsubset> src_net \<and> dest pa \<sqsubset> dest_net}"

definition 
  deny_all_from_to    :: "'\<alpha>::adr net \<Rightarrow> '\<alpha>::adr net \<Rightarrow> (('\<alpha>,'\<beta>) packet \<mapsto> unit)" where
  "deny_all_from_to src_net dest_net = deny_all |`
                                   {pa. src pa \<sqsubset> src_net \<and> dest pa \<sqsubset> dest_net}"






consts noTranslation :: "(('\<alpha>,'\<beta>) packet \<mapsto> unit) \<Rightarrow> ('\<alpha>,'\<beta>) packet \<mapsto> ('\<alpha>,'\<beta>) packet"

definition idPolicy :: "('\<alpha> \<mapsto> unit) \<Rightarrow> ('\<alpha> \<mapsto> '\<alpha>)" where
 "idPolicy p = (\<lambda> x. (case p x of Some (allow ()) \<Rightarrow> Some (allow x)
                             |Some (deny ()) \<Rightarrow> Some (deny x)
                             | None \<Rightarrow> None))"


lemma "(allow_pfun (idPolicy p)) = s"
oops

definition 
  allow_allO    :: "('\<alpha>,'\<beta>) packet \<mapsto> ('\<alpha>,'\<beta>) packet" where 
  "allow_allO  = idPolicy allow_all" 


definition 
  allow_all1    :: "('\<alpha>,'\<beta>) packet \<mapsto> ('\<alpha>,'\<beta>) packet" where 
  "allow_all1 p  = Some (allow p)" 

lemma "allow_allO = allow_all1"
apply (rule ext)
apply (simp add: allow_all1_def allow_allO_def)
apply (simp add: idPolicy_def allow_all_def)
apply (simp split: option.splits unit.splits)
done



text{* All these combinators and the default rules are put into one
  single lemma called @{text PolicyCombinators} to faciliate proving
  over policies.  *}


lemmas PolicyCombinators =
  allow_all_def deny_all_def allow_all_from_def deny_all_from_def
  allow_all_to_def deny_all_to_def allow_all_from_to_def deny_all_from_to_def
  map_add_def restrict_map_def idPolicy_def
                         
end
