(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * AVL_test_SMT.thy --- test data generation for AVL trees using the Z3 solver
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007 ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: AVL_test.thy 8951 2010-05-18 15:02:26Z makarius $ *)

header {* Testing AVL Properties *}

theory 
  AVL_test_SMT
imports 
  AVL_def
begin

declare [[testgen_profiling]]
declare [[testgen_SMT]]
declare [[ smt_solver = z3 ]]

text {* This test plan of the theory follows more or less the
        standard. However, we insert some minor theorems
        into the test theorem generation in order to
        ease the task of solving; this both improves speed
        of the generation and quality of the test.  *}

declare insert_base insert_rec [simp del]

lemma size_0[simp]: "(size x = 0) = (x = ET)"
  by(induct "x",auto)

lemma height_0[simp]: "(height x = 0) = (x = ET)";
  by(induct "x",auto) 

lemma [simp]: "(max (Suc a) b) ~= 0"
  by(auto simp: max_def)

lemma [simp]: "(max b (Suc a) ) ~= 0"
  by(auto simp: max_def)

text {* We adjust the random generator to a fairly restricted
        level and go for the solving phase. *}

declare [[testgen_iterations=10]]

test_spec "(is_bal t) --> (is_bal (insert  x t))"
  apply(gen_test_cases "insert")
  store_test_thm "foo"

lemma tree_exhaust: "trigger [[pat (size t)]] (t = ET \<or> (\<exists> a b c. t = MKT a b c))"
apply(simp add: trigger_def)
apply(rule_tac y = "t" in  tree.exhaust)
apply(auto)
done

schematic_lemma tree_inject: "trigger [[pat(MKT ?a ?tree1.0 ?tree2.0),pat(MKT ?a' ?tree1' ?tree2')]] 
((MKT ?a ?tree1.0 ?tree2.0 = MKT ?a' ?tree1' ?tree2') =
(?a = ?a' \<and> ?tree1.0 = ?tree1' \<and> ?tree2.0 = ?tree2'))"
apply(simp add: trigger_def)
done

schematic_lemma tree_distinct: "trigger [[pat(MKT ?a' ?tree1' ?tree2')]] (ET \<noteq> MKT ?a' ?tree1' ?tree2')"
apply(simp add: trigger_def)
done

declare is_bal.simps [testgen_smt_facts]
declare height.simps [testgen_smt_facts]
(* declare tree_inject [testgen_smt_facts]
declare tree_distinct [testgen_smt_facts]
declare tree.size(3) [testgen_smt_facts]
declare tree.size(4) [testgen_smt_facts]
declare tree_exhaust [testgen_smt_facts] *)

thm testgen_smt_facts

declare [[ z3_options = "AUTO_CONFIG=false" ]]
declare [[smt_datatypes=true]]

  gen_test_data "foo"

thm foo.test_data

end
