(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * isar_setup.sml --- Isar setup for HOL-TestGen
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007 ETH Zurich, Switzerland
 *               2008 Achim D. Brucker, Germany
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: isar_setup.ML 8955 2010-05-27 16:04:13Z makarius $ *)

(* at the moment, calling store_test_thm without the optional argument  works only 
   in an ProofGeneral session because otherweise the call of Toplevel.pretty_state 
   results in an empty list \<dots> 
*)
structure P = OuterParse and K = OuterKeyword;

			     
fun store_test_thm name = Toplevel.end_proof (fn _ => fn state =>
  let
    val ctxt = Proof.context_of state;
    val _ = if Config.get ctxt TestGen.profiling then 
		(Clocks.start_clock "unnamed_test_thm"; 
		 Clocks.rename_clock "unnamed_test_thm" name; 
		 Clocks.stop_clock name) 
	    else 
		()

    val _ = Proof.assert_backward state;
    val {context = goal_ctxt, goal, ...} = Proof.goal state
    fun up_thy thy = 
	let 	  
	  fun add_to_dm name thm thy = 
	      let
		val thy = TestEnv.map_data_global (TestEnv.add_test_case (name,thm)) thy
                val t = LogThy.get_tc_delta ()
		val num_of_tc = List.length 
				(List.filter (fn tc => not (TestGen.is_thyp tc)) (prems_of thm))
		val _ = writeln (concat ["Test theorem (gen_test_cases) '",name,"': ", 
				       Int.toString num_of_tc, 
				       " test cases in ",Time.toString t, " seconds"])

		val _ = LogThy.append (concat [Context.theory_name thy, ", "
				    ,name, ", " 
				    ,"test case, " 
				    ,Int.toString num_of_tc,", " 
				    , Time.toString t,"\n"])
              in
		thy
	      end
	in
	  thy |> Sign.add_path (space_implode "_" [name])
	      |> (PureThy.add_thms [((@{binding test_thm}, goal), [])])
	      |> snd
	      |> Sign.parent_path
	      |> add_to_dm (space_implode "_" [name]) goal
	      |> Clocks.flush_clocks
	end

  in goal_ctxt |> Local_Theory.theory up_thy  end);

val _ =
  OuterSyntax.command "store_test_thm" "store test state (theorem)"
    (OuterKeyword.tag_proof OuterKeyword.qed)        
    (OuterParse.name >> store_test_thm);   




    
(**********************)
fun gen_test_dataT name thy = 
    let 
	fun gen_test_data name   = 
	    let
		val _  = LogThy.start_td_timer ()
		val _ = if Config.get_thy thy TestGen.profiling then 
			    (Clocks.init_clocks thy;
			     Clocks.start_clock name;
			     Clocks.start_clock "gen_test_data")
			else 
			    ()
		
		val thy = Context.theory_map (TestGen.gen_test_data name) thy;
                val t = LogThy.get_td_delta ()
		val thy = if Config.get_thy thy TestGen.profiling then 
			      (Clocks.stop_clock "gen_test_data";
			       Clocks.stop_clock name;
			       Clocks.flush_clocks thy)
			  else 
			      thy

		val dm = TestEnv.get_data_global thy;
		val data = TestGen.get_test_data (dm,name)
		val hyps = TestGen.get_test_hyps (dm,name); 
		val thy  = Sign.add_path (space_implode "_" [name]) thy;
                val thy = snd(PureThy.add_thmss [((@{binding test_hyps},hyps),[])] (thy)); 
		val thy = snd(PureThy.add_thmss [((@{binding test_data},data),[])] (thy));
		val _ = writeln (concat ["Test theorem (gen_test_data) '",name,"': ", 
				       Int.toString (List.length data), 
				       " test cases in ",Time.toString t, " seconds"])

		val _ = LogThy.append (concat [Context.theory_name thy, ", "
				    ,name, ", " 
				    ,"test data, "
				    ,Int.toString (List.length data),", " 
				    ,Time.toString t,"\n"])

                val thy = Sign.parent_path thy;
	    in 
		thy
	    end
        val thy = gen_test_data name 
    in 
        thy
    end 
    
val _ =
    OuterSyntax.command "gen_test_data" "generate test data" 
			OuterKeyword.thy_script
			((P.name)  >> (Toplevel.theory o gen_test_dataT));
    
(**********************)

val _ =
    OuterSyntax.command "gen_test_script" "generate test script" 
			OuterKeyword.thy_script
			((P.name  -- P.name  -- P.name -- Scan.optional P.name "")
				>> (fn (((filename,name),prog),smlprogname) =>
					(Toplevel.theory (TestGen.gen_testscript prog smlprogname filename name))));
(**********************)

(**********************)
	
val _ =
    OuterSyntax.command "export_test_data" "export test data to an external file" 
			OuterKeyword.thy_script
			((P.name  -- P.name)
				>> (fn (filename,name) =>
					(Toplevel.theory (TestEnv.export_test_data filename name))));
(**********************)
    

val _ =
  OuterSyntax.local_theory_to_proof "test_spec" "define test specification" K.thy_goal
    (Scan.optional (SpecParse.opt_thm_name ":" --|
      Scan.ahead (SpecParse.locale_keyword || SpecParse.statement_keyword)) Attrib.empty_binding --
      SpecParse.general_statement >> (fn (a, (elems, concl)) => fn lthy =>
          let val _ = if Config.get lthy TestGen.profiling then Clocks.init_clocks (ProofContext.theory_of lthy) else ()
          in
              Specification.theorem_cmd "test_spec" NONE (K I) a elems concl false lthy
	  end));



fun update_typlist sl thy (x:TestEnv.testenv) = 
let
    val typlist =  map (fn ts => (Syntax.read_typ (ProofContext.init thy) ts)) sl;
in 
    TestEnv.type_candiates_update typlist x
end

fun app [] x = x
  | app (f :: fs) x = app fs (f x);


val testgen_params =
  [("type_candidates",(P.$$$ "[" |-- P.list1 P.typ --| P.$$$ "]") >> update_typlist)];



val parse_testgen_params = P.short_ident :-- (fn s =>
  P.$$$ "=" |-- the_default Scan.fail (AList.lookup (op =) testgen_params s)) >> snd;

val _ =
  OuterSyntax.command "testgen_params" "set parameters for TestGen" K.thy_decl
    (P.$$$ "[" |-- P.list1 parse_testgen_params --| P.$$$ "]" >> 
     (fn fs => Toplevel.theory (fn thy =>
         TestEnv.map_data_global (app (map (fn f => f (thy )) fs)) thy)));
